/*
** Module   : TableResultSet.java
** Abstract : A result set used to serialize a table data and send it to a remote side. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130610 Created initial version.
** 002 CA  20130628 Added possibility to retrieve the data as a ResultSet instance.
** 003 OM  20130705 Fixed date/datetime results with time functions.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.math.BigDecimal;
import java.sql.*;
import java.util.*;
import org.hibernate.usertype.UserType;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;

/**
 * Provides the main mechanism of passing table-like metadata and data which the remote side will
 * interpret as a temporary table. The remote side will use the sent properties (and their other
 * settings, as type and extent) to dynamically build the DMO interface, implementation class and
 * .hbm.xml configuration file.
 * <p>
 * As transport, any instance of this class will be wrapped in a {@link TableWrapper} instance
 * internally by P2J.
 */
public abstract class TableResultSet
{
   /** Flag indicating this table is sent in INPUT or INPUT-OUTPUT mode. */
   private boolean input;

   /** Flag indicating this table is sent in OUTPUT or INPUT-OUTPUT mode. */
   private boolean output;
   
   /** Flag indicating this table is sent in APPEND mode. */
   private boolean append;
   
   /**
    * The data sent by a remote side, after the remote appserver call, when this table is in 
    * OUTPUT or INPUT-OUTPUT mode.
    */
   private List<Object[]> rows;
   
   /**
    * Default c'tor, explicitly added to allow instances of this class to be created on 
    * deserialization. Not for public use.
    */
   public TableResultSet()
   {
   }
   
   /**
    * Create a new instance of this result set.
    * 
    * @param    input
    *           Flag indicating this table is sent in INPUT or INPUT-OUTPUT mode.
    * @param    output
    *           Flag indicating this table is sent in OUTPUT or INPUT-OUTPUT mode.
    * @param    append
    *           Flag indicating this table is sent in APPEND mode.
    */
   public TableResultSet(boolean input, boolean output, boolean append)
   {
      this.input = input;
      this.output = output;
      this.append = append;
   }

   /**
    * Used in conjunction with {@link #nextProperty()}, it provides a mechanism for subclasses
    * to send the property list in an iterator-like form.
    * <p>
    * Override this method so that it returns <code>false</code> when {@link #nextProperty} has
    * sent all properties.
    * 
    * @return   Default it returns <code>false<code>.
    */
   public abstract boolean hasMoreProperties();
   
   /**
    * Used in conjunction with {@link #hasMoreProperties()}, it provides a mechanism for 
    * subclasses to send the property list in an iterator-like form.
    * <p>
    * Override this method so that when {@link #hasMoreProperties} returns <code>true</code>, this
    * will return the name of the next property in the list.
    * 
    * @return   Default it returns <code>null<code>.
    */
   public abstract PropertyDefinition nextProperty();
   
   /**
    * Used in conjunction with {@link #hasMoreRows()}, it provides a mechanism for subclasses to
    * send the data in an iterator-like form.
    * <p>
    * Override this method so that it returns <code>false</code> when {@link #nextRow} has sent 
    * all data.
    * 
    * @return   Default it returns <code>false<code>.
    */
   public abstract boolean hasMoreRows();
   
   /**
    * Used in conjunction with {@link #hasMoreRows()}, it provides a mechanism for  subclasses to
    * send the data in an iterator-like form.
    * <p>
    * Override this method so that when {@link #hasMoreRows} returns <code>true</code>, this will
    * return the next row in the table.
    * 
    * @return   Default it returns <code>null<code>.
    */
   public abstract Object[] nextRow();
   
   /**
    * Set the data sent by a remote side, after the remote appserver call.
    * 
    * @param    rows
    *           The data sent by the remote side.
    */
   public void setRows(List<Object[]> rows)
   {
      this.rows = rows;
   }
   
   /**
    * Get the data sent by a remote side, after the remote appserver call. This will be non-null
    * only if the parameter is in OUTPUT or INPUT-OUTPUT mode.
    * 
    * @return   See above.
    */
   public List<Object[]> getRows()
   {
      return rows;
   }
   
   /**
    * Check if this table is in INPUT or INPUT-OUTPUT mode.
    * 
    * @return   The {@link #input} flag.
    */
   public boolean isInput()
   {
      return input;
   }
   
   /**
    * Check if this table is in OUTPUT or INPUT-OUTPUT mode.
    * 
    * @return   The {@link #output} flag.
    */
   public boolean isOutput()
   {
      return output;
   }
   
   /**
    * Check if this table is in APPEND mode.
    * 
    * @return   The {@link #append} flag.
    */
   public boolean isAppend()
   {
      return append;
   }
   
   /** 
    * Get the data sent by a remote side, after the remote appserver call, as a {@link ResultSet}
    * instance.  This will be non-null only if the parameter is in OUTPUT or INPUT-OUTPUT mode.
    * 
    * @return   See above.
    */
   public ResultSet asResultSet()
   {
      if (!isOutput())
      {
         return null;
      }
      
      Class<?>[] interfaces = { ResultSet.class };
      
      ResultSet proxy = (ResultSet) ProxyFactory.getProxy(ResultSet.class,
                                                          interfaces,
                                                          false,
                                                          null,
                                                          new DataHandler());
      
      return proxy;
   }

   /**
    * Handler for the {@link ResultSetMetaData} proxy.
    */
   private final class MetaDataHandler
   implements InvocationHandler
   {
      /** The property list. */
      private final List<PropertyDefinition> props;
      
      /**
       * Build a new instance which extracts metadata from the given property list.
       * 
       * @param    props
       *           The property list.
       */
      public MetaDataHandler(List<PropertyDefinition> props)
      {
         this.props = props;
      }
      
      /**
       * Handler for {@link ResultSetMetaData} APIs. Only the following are supported:
       * <ul>
       *    <li>{@link ResultSetMetaData#getColumnCount}</li>
       *    <li>{@link ResultSetMetaData#getColumnName}</li>
       *    <li>{@link ResultSetMetaData#getColumnTypeName}</li>
       *    <li>{@link ResultSetMetaData#getColumnType}</li>
       * </li>
       */
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         String mname = method.getName();
         
         if ("getColumnCount".equals(mname))
         {
            return props.size();
         }
         
         if (mname.startsWith("get") && args.length == 1 && args[0] instanceof Integer)
         {
            int idx = (Integer) args[0] - 1;
            PropertyDefinition pd = props.get(idx);

            if ("getColumnName".equals(mname))
            {
               return pd.getName();
            }
            
            if ("getColumnTypeName".equals(mname))
            {
               return pd.getType().getName();
            }
            
            if ("getColumnType".equals(mname))
            {
               UserType ut = pd.getType().newInstance();
               return ut.sqlTypes()[0];
            }
         }
         
         throw new SQLFeatureNotSupportedException("Method " + mname + " is not supported!");
      }
   }

   /**
    * Handler for the {@link ResultSet} proxy.  This will provide the data in a 
    * {@link ResultSet#TYPE_FORWARD_ONLY forward-only} manner.
    */
   private final class DataHandler
   implements InvocationHandler
   {
      /** The {@link ResultSetMetaData} proxy. */
      private final ResultSetMetaData rsmd;

      /** The list of properties. */
      private final List<PropertyDefinition> props;

      /** An iterator over the data. */
      private Iterator<Object[]> iter;
      
      /** Flag indicating if the last retrieved column was SQL-null. */
      private boolean wasNull = false;
      
      /** The data for the current row. */
      private Object[] data = null;
      
      /**
       * Create a new instance. This initializes the iterator, builds the 
       * {@link #props property list} and the {@link #rsmd metadata}.
       */
      public DataHandler()
      {
         iter = rows.iterator();

         props = new ArrayList<PropertyDefinition>();
         while (hasMoreProperties())
         {
            props.add(nextProperty());
         }

         Class<?>[] interfaces = { ResultSetMetaData.class };
         rsmd = (ResultSetMetaData) ProxyFactory.getProxy(ResultSetMetaData.class,
                                                          interfaces,
                                                          false,
                                                          null,
                                                          new MetaDataHandler(props));
      }
      
      /**
       * Handler for the {@link ResultSet} APIs. Only the following are supported:
       * <ul>
       *    <li>{@link ResultSet#next}</li>
       *    <li>{@link ResultSet#close}</li>
       *    <li>{@link ResultSet#isClosed}</li>
       *    <li>{@link ResultSet#wasNull}</li>
       *    <li>{@link ResultSet#findColumn}</li>
       *    <li>{@link ResultSet#relative}</li>
       *    <li>{@link ResultSet#getType}</li>
       *    <li>{@link ResultSet#getMetaData}</li>
       *    <li>{@link ResultSet#getObject}</li>
       *    <li>{@link ResultSet#getBoolean}</li>
       *    <li>{@link ResultSet#getDouble}</li>
       *    <li>{@link ResultSet#getBigDecimal}</li>
       *    <li>{@link ResultSet#getFloat}</li>
       *    <li>{@link ResultSet#getInt}</li>
       *    <li>{@link ResultSet#getLong}</li>
       *    <li>{@link ResultSet#getShort}</li>
       *    <li>{@link ResultSet#getByte}</li>
       *    <li>{@link ResultSet#getString}</li>
       *    <li>{@link ResultSet#getDate}</li>
       *    <li>{@link ResultSet#getTime}</li>
       *    <li>{@link ResultSet#getTimestamp}</li>
       * </ul>
       */
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         String mname = method.getName();

         if ("next".equals(mname))
         {
            boolean next = iter.hasNext();
            if (next)
            {
               data = iter.next();
            }
            
            return next;
         }

         if ("close".equals(mname))
         {
            iter = null;
            return null;
         }
         
         if ("isClosed".equals(mname))
         {
            return (iter == null);
         }

         if ("wasNull".equals(mname))
         {
            return wasNull;
         }
         
         if ("findColumn".equals(mname))
         {
            return getPropertyIndex(mname);
         }
         
         if ("relative".equals(mname))
         {
            int rel = ((Integer) args[0]).intValue();
            if (rel < 0)
            {
               throw new SQLFeatureNotSupportedException("The result set is forward only!");
            }
            else
            {
               for (int i = 0; i < rel && iter.hasNext(); i++)
               {
                  data = iter.next();
               }
            }
            
            return null;
         }

         if ("getType".equals(mname))
         {
            return ResultSet.TYPE_FORWARD_ONLY;
         }
         
         if ("getMetaData".equals(mname))
         {
            return rsmd;
         }
         
         Object res = null;
         if (mname.startsWith("get"))
         {
            wasNull = false;

            int idx = 0;
            if (args[0] instanceof String)
            {
               idx = getPropertyIndex((String) args[0]);
            }
            else
            {
               idx = (Integer) args[0];
            }

            // make this 0-based index
            idx = idx - 1;

            Object val = data[idx];
            boolean resolved = true;
            boolean npe = false;

            if ("getObject".equals(mname))
            {
               res = val;
            }
            else if ("getBoolean".equals(mname))
            {
               npe = true;

               if (val instanceof Boolean)
               {
                  res = ((Boolean) val).booleanValue();
               }
               else if (val instanceof logical)
               {
                  res = ((logical) val).booleanValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getDouble".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).doubleValue();
               }
               else if (val instanceof NumberType)
               {
                  res = ((NumberType) val).doubleValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getBigDecimal".equals(mname))
            { 
               if (val instanceof Number)
               {
                  res = new BigDecimal(((Number) val).doubleValue());
               }
               else if (val instanceof NumberType)
               {
                  res = new BigDecimal(((NumberType) val).doubleValue());
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getFloat".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).floatValue();
               }
               else if (val instanceof NumberType)
               {
                  res = (float) ((NumberType) val).doubleValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getInt".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).intValue();
               }
               else if (val instanceof NumberType)
               {
                  res = ((NumberType) val).intValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getLong".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).longValue();
               }
               else if (val instanceof NumberType)
               {
                  res = ((NumberType) val).longValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getShort".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).shortValue();
               }
               else if (val instanceof NumberType)
               {
                  res = (short) ((NumberType) val).intValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getByte".equals(mname))
            {
               npe = true;

               if (val instanceof Number)
               {
                  res = ((Number) val).byteValue();
               }
               else if (val instanceof NumberType)
               {
                  res = (byte) ((NumberType) val).intValue();
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getString".equals(mname))
            {
               if (val instanceof BaseDataType)
               {
                  res = ((BaseDataType) val).toStringMessage();
               }
               else 
               {
                  res = (val == null ? null : val.toString());
               }
            }
            else if ("getDate".equals(mname))
            {
               // TODO: Calendar parameter?
               if (val instanceof java.sql.Date)
               {
                  res = val;
               }
               else if (val instanceof java.util.Date)
               {
                  res = new java.sql.Date(((java.util.Date) val).getTime());
               }
               else if (val instanceof date)
               {
                  res = new java.sql.Date(((date) val).dateValue().getTime());
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getTime".equals(mname))
            {
               // TODO: Calendar parameter?
               if (val instanceof java.sql.Time)
               {
                  res = val;
               }
               else if (val instanceof java.util.Date)
               {
                  res = new java.sql.Time(((java.util.Date) val).getTime());
               }
               else if (val instanceof date)
               {
                  res = new java.sql.Time(((date) val).dateValue().getTime());
               }
               else
               {
                  resolved = false;
               }
            }
            else if ("getTimestamp".equals(mname))
            {
               // TODO: Calendar parameter?
               if (val instanceof java.sql.Timestamp)
               {
                  res = val;
               }
               else if (val instanceof java.util.Date)
               {
                  res = new java.sql.Timestamp(((java.util.Date) val).getTime());
               }
               else if (val instanceof date)
               {
                  res = new java.sql.Timestamp(((date) val).dateValue().getTime());
               }
               else
               {
                  resolved = false;
               }
            }
            
            if (npe)
            {
               throw new NullPointerException("Value for column " + args[0].toString() + 
                                              " is null, when calling " + mname + "!");
            }
            else if (!resolved)
            {
               throw new SQLException("Value for column " + args[0].toString() + 
                                      " is not compatible with mehod " + mname + "!");
            }
            else
            {
               wasNull = (res != null);
            }
            
            return res;
         }
         
         throw new SQLFeatureNotSupportedException("Method " + mname + " is not supported!");
      }

      /**
       * Get the 1-based index for the property with the specified name. Search is done 
       * case-insensitive.  If not found, an {@link SQLException} is throw.
       * 
       * @param    name
       *           The property name.
       *           
       * @return   The 1-based index in the property list.
       * 
       * @throws   SQLException
       *           If the property was not found.
       */
      private int getPropertyIndex(String name)
      throws SQLException
      {
         for (int i = 0; i < props.size(); i++)
         {
            PropertyDefinition pd = props.get(i);
            if (pd.getName().equalsIgnoreCase(name))
            {
               return i + 1;
            }
         }
            
         throw new SQLException("Property " + name + " was not found!");
      }
   }
}
