/*
** Module   : Int64UserType.java
** Abstract : Custom user type to wrap P2J int64 for Hibernate access
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20130121 Created initial version. Maps P2J int64 to SQL BIGINT.
** 002 SVL 20120331 Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;
import org.hibernate.engine.spi.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.int64 P2J int64}. This permits Hibernate to
 * convert this data type to and from the SQL BIGINT data type. The unknown
 * value is mapped to SQL NULL.
 */
public final class Int64UserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { Types.BIGINT };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map an int64 to/from a database
    * record.
    *
    * @return  Array containing a single type constant:  BIGINT.
    */
   @Override
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.int64 P2J int64} class.
    */
   public Class returnedClass()
   {
      return int64.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>int64</code> instance.
    *
    * @param   value
    *          The <code>int64</code> instance to be copied.
    *
    * @return  A new instance of <code>int64</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      NumberType num = (NumberType) value;
      
      if (num.isUnknown())
      {
         return (new int64());
      }
      
      return (new int64(num.longValue()));
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return an <code>int64</code>
    * instance which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>long</code>.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   session
    *          Not used.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  An <code>int64</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs,
                             String[] names,
                             SessionImplementor session,
                             Object owner)
   throws SQLException
   {
      long value = rs.getLong(names[0]);
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.trace("returning unknown value as column: " + names[0]);
         }
         return new int64();
      }
      
      int64 wrapper = new int64(value);
      if (LOG_TRACE)
      {
         LOG.trace("returning '" + toString(wrapper) + "' as column: " +
                   names[0]);
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>long</code> value into the given prepared statement,
    * based upon the <code>int64</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          An <code>int64</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType      value,
                              int               index)
   throws SQLException
   {
      ps.setLong(index, ((NumberType) value).longValue());
   }
   
   /**
    * Return the string value of the <code>int</code> embedded in the
    * given <code>integer</code> object <code>value</code>. The default
    * <code>toString()</code> method of <code>integer</code> returns a
    * formatted, padded string, which is not suitable for our purposes.
    *
    * @param   value
    *          An <code>int64</code> object. It is safe to assume this
    *          will not be <code>null</code>.
    *
    * @return  The <code>String</code> representation of the contents of
    *          <code>value</code>.
    */
   protected String toString(BaseDataType value)
   {
      return String.valueOf(((NumberType) value).longValue());
   }
}
